/*
	CBerkeleyDB.cpp
	Classe base per interfaccia con Berkeley DB (usare solo come virtuale pura).
	Luca Piergentili, 04/11/99
	lpiergentili@yahoo.com
	http://www.geocities.com/lpiergentili/
*/
#include "env.h"
#include "pragma.h"
#include "macro.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "strcpyn.h"
#include <errno.h>
#ifdef _WINDOWS
  #include "window.h"
  #ifdef WIN32_MFC
    #include "CMutexEx.h"
  #endif
#else
  #include "typedef.h"
#endif
#include "traceexpr.h"
#include "db.h"
#include "CBerkeleyDB.h"

// array per codici d'errore
struct DB_ERROR {
	int		code;
	char*	desc;
};

static const DB_ERROR db_error[] = {
	// 1000 (info)
	DB_OK,						"Ok",
	// 2000 (status)
	DB_RETCODE_NOTFOUND,			"Not found",
	DB_RETCODE_BOF,				"Bof",
	DB_RETCODE_EOF,				"Eof",
	// 3000 (errori)
	DB_RETCODE_EOPENTABLE,			"Error opening table",
	DB_RETCODE_EOPENCURSOR,			"Error opening cursor",
	DB_RETCODE_EOPENINDEX,			"Error opening index",
	DB_RETCODE_EOPENIDXCURSOR,		"Error opening index cursor",
	DB_RETCODE_ECREATETABLE,			"Error creating table",
	DB_RETCODE_ECLOSETABLE,			"Error closing table",
	DB_RETCODE_ECLOSECURSOR,			"Error closing cursor",
	DB_RETCODE_ECLOSEINDEX,			"Error closing index",
	DB_RETCODE_ECLOSEIDXCURSOR,		"Error closing index cursor",
	DB_RETCODE_EALREDYOPEN,			"Table is alredy open",
	DB_RETCODE_EINVALIDHANDLE,		"Invalid handle",
	DB_RETCODE_ELOCKFAILURE,			"Lock failure",
	DB_RETCODE_EINVALIDPARAM,		"Invalid parameter",
	DB_RETCODE_EINVALIDRESOURCE,		"Invalid resource",
	DB_RETCODE_EINVALIDINDEX,		"Invalid index",
	DB_RETCODE_ETABLENOTOPEN,		"Table not open",
	DB_RETCODE_EINVALIDFIELDNUMBER,	"Invalid field number",
	DB_RETCODE_EINVALIDFIELDSIZE,		"Invalid field size",
	DB_RETCODE_EALLOCFAILURE,		"Memory alloc failure",
	DB_RETCODE_EOPENFAILURE,			"Open failure",
	// >3000 (errori)
	DB_RETCODE_ERROR,				"Generic error",
	DB_RETCODE_UNKNOWERROR,			"Unknow error"
};

/*
	CBerkeleyDB()
*/
CBerkeleyDB::CBerkeleyDB()
{
	memset(&m_Database,'\0',sizeof(DATABASE));
	m_pDataBuffer = NULL;
	m_nDataBufferSize = -1;
	m_nRecordSize = -1;
	memset(&m_dbenv,'\0',sizeof(DB_ENV));
	memset(&m_dbinfo,'\0',sizeof(DB_INFO));
	memset(&m_Key,'\0',sizeof(DBT));
	memset(&m_Data,'\0',sizeof(DBT));
	m_pDbHandle = NULL;
	m_pDbCursor = NULL;
	m_pCurrentCursor = NULL;
	m_nCurrentCursorIndex = -1;
	m_pPrimaryKeyCursor = NULL;
	m_nPrimaryKeyCursorIndex = -1;
	m_pIdxHandleArray = NULL;
	m_pIdxCursorArray = NULL;
	memset(m_szLastKey,'\0',sizeof(m_szLastKey));
#if defined(_WINDOWS)
	m_bShowErrors = FALSE;
#endif
}

/*
	~CBerkeleyDB()
*/
CBerkeleyDB::~CBerkeleyDB()
{
	Close();
}

/*
	Open()

	Apre la tabella, i cursori e gli indici associati.
	Puo' essere chiamata piu' volte, per cui predisporre le inizializzazioni in modo adeguato (non assumere
	una sola chiamata, inizializzando/rilasciando solo nel costruttore/distruttore).
	Non effettua test sul risultato delle new [].

	Se la tabella non prevede indici, il primo campo viene considerato dall'api come chiave, per cui la 
	tabella viene impostata per contenere chiavi duplicate dato che sicuramente conterra' chiavi doppie.
	Se la tabella prevede indici, il primo campo, dato che viene considerato dall'api come chiave, viene
	utilizzato internamente come chiave primaria per cui non viene impostato il flag per le chiavi duplicate.
	
	Nel primo caso la ricerca deve essere effettuata leggendo un record alla volta in un ciclo e confrontando
	il valore del record.
	Nel secondo, si seleziona l'indice e si effettua una seek. i campi utilizzati come chiavi non possono
	contenere duplicati
*/
int CBerkeleyDB::Open(u_int32_t flags/*=DB_CREATE*/)
{
	int i,nCode = 0;

	// se la tabella non prevede indici il primo campo viene considerato come chiave per cui
	// deve impostare il flag per poter inserire record duplicati
	memset(&m_dbinfo,'\0',sizeof(DB_INFO));
	m_dbinfo.flags = (m_Database.table.totindex > 0) ? 0 : DB_DUP|DB_DUPSORT;

	// apre (o crea) la tabella
#if 0
	memset(&m_dbenv,'\0',sizeof(DB_ENV));
	db_appinit("Drive:\\DatabasePathname",NULL,&m_dbenv,DB_CREATE|DB_INIT_LOCK);
	if((nCode = db_open(m_Database.table.filename,DB_BTREE,flags,0664,&m_dbenv,&m_dbinfo,&m_pDbHandle))!=0)
#else
	if((nCode = db_open(m_Database.table.filename,DB_BTREE,flags,0664,NULL,&m_dbinfo,&m_pDbHandle))!=0)
#endif
	{
		m_pDbHandle = NULL;
		SetLastError(DB_RETCODE_EOPENTABLE);
		goto done;
	}
	// cursore associato
	if((nCode = m_pDbHandle->cursor(m_pDbHandle,NULL,&m_pDbCursor,0))!=0)
	{
		m_pDbCursor = NULL;
		SetLastError(DB_RETCODE_EOPENCURSOR);
		goto done;
	}
	
	m_pCurrentCursor        = m_pDbCursor;
	m_nCurrentCursorIndex    = -1;
	m_pPrimaryKeyCursor     = m_pDbCursor;
	m_nPrimaryKeyCursorIndex = -1;

	// alloca il buffer per il record (somma della dimensione dei campi)
	for(i = 0,m_nRecordSize = 0; i < m_Database.table.totfield; i++)
		m_nRecordSize += m_Database.table.row[i].size;
	
	ASSERTEXPR(m_nRecordSize > 0 && m_nRecordSize <= (MAX_FIELDSIZE*MAX_FIELDCOUNT));
	
	if(m_nRecordSize > 0)
	{
		m_nDataBufferSize = m_nRecordSize + 1;
		m_pDataBuffer = new char[m_nDataBufferSize];
		ASSERTEXPR(m_pDataBuffer!=NULL);
		memset(m_pDataBuffer,'\0',m_nDataBufferSize);
	}
	else
	{
		nCode = SetLastError(DB_RETCODE_EINVALIDFIELDSIZE);
		goto done;
	}

	// se la tabella prevede indici
	if(m_Database.table.totindex > 0)
	{
		// ricava l'ultima chiave (progressivo numerico)
		// necessario per le eventuali successive chiamate a Insert()
		ResetPair();
		memset(m_szLastKey,'\0',sizeof(m_szLastKey));
		if(m_pDbCursor->c_get(m_pDbCursor,&m_Key,&m_Data,DB_LAST)==0)
			memcpy(m_szLastKey,m_Key.data,GetPrimaryKeySize());
		else
			m_szLastKey[0] = '0';
		ResetPair();

		// alloca gli array per gli handles ed i cursori degli indici
		m_pIdxHandleArray = (DB**)new DB[m_Database.table.totindex];
		ASSERTEXPR(m_pIdxHandleArray!=NULL);
		m_pIdxCursorArray = (DBC**)new DBC[m_Database.table.totindex];
		ASSERTEXPR(m_pIdxCursorArray!=NULL);

		// apre (o crea) gli indici
		m_dbinfo.flags = DB_DUP|DB_DUPSORT;
		for(i = 0; i < m_Database.table.totindex; i++)
		{
			// indice
			if((nCode = db_open(m_Database.table.index[i].filename,DB_BTREE,flags,0664,NULL,&m_dbinfo,&m_pIdxHandleArray[i]))!=0)
			{
				m_pIdxHandleArray[i] = NULL;
				SetLastError(DB_RETCODE_EOPENINDEX);
				goto done;
			}
			// cursore
			if((nCode = m_pIdxHandleArray[i]->cursor(m_pIdxHandleArray[i],NULL,&m_pIdxCursorArray[i],0))!=0)
			{
				m_pIdxCursorArray[i] = NULL;
				SetLastError(DB_RETCODE_EOPENIDXCURSOR);
				goto done;
			}
		}
	}
	else
	{
		m_pIdxHandleArray = NULL;
		m_pIdxCursorArray = NULL;
	}

	// azzera il record
	ResetRecord();

done:

	// errore
	if(nCode!=0)
		Close();

	return(nCode);
}

/*
	Close()

	Chiude la tabella, i cursori e gli indici associati.
	Puo' essere chiamata piu' volte, per cui predisporre le inizializzazioni in modo adeguato (non assumere
	una sola chiamata, inizializzando/rilasciando solo nel costruttore/distruttore).
*/
int CBerkeleyDB::Close(void)
{
	int nCode = 0;

	// chiude i cursori degli indici
	if(m_Database.table.totindex > 0)
	{
		if(m_pIdxCursorArray!=NULL)
			for(int i = 0; i < m_Database.table.totindex; i++)
			{
				if(m_pIdxCursorArray[i]!=NULL)
					if(m_pIdxCursorArray[i]->c_close(m_pIdxCursorArray[i])!=0)
						nCode = SetLastError(DB_RETCODE_ECLOSEIDXCURSOR);
			}
	}

	// chiude il cursore della tabella
	if(m_pDbCursor!=NULL)
		if(m_pDbCursor->c_close(m_pDbCursor)!=0)
			nCode = SetLastError(DB_RETCODE_ECLOSECURSOR);
	
	// chiude gli indici
	if(m_Database.table.totindex > 0)
	{
		if(m_pIdxHandleArray!=NULL)
			for(int i = 0; i < m_Database.table.totindex; i++)
			{
				if(m_pIdxHandleArray[i]!=NULL)
					if(m_pIdxHandleArray[i]->close(m_pIdxHandleArray[i],0)!=0)
						nCode = SetLastError(DB_RETCODE_ECLOSEINDEX);
			}
	}

	// chiude la tabella
	if(m_pDbHandle!=NULL)
		if(m_pDbHandle->close(m_pDbHandle,0)!=0)
			nCode = SetLastError(DB_RETCODE_ECLOSETABLE);

	m_pDbHandle = NULL;
	m_pDbCursor = NULL;

	// rilascia il buffer per il record
	if(m_pDataBuffer)
		delete [] m_pDataBuffer,m_pDataBuffer = NULL;

	// rilascia gli array per gli indici
	if(m_Database.table.totindex > 0)
	{
		if(m_pIdxHandleArray)
			delete [] m_pIdxHandleArray,m_pIdxHandleArray = NULL;
		if(m_pIdxCursorArray)
			delete [] m_pIdxCursorArray,m_pIdxCursorArray = NULL;
	}

	return(nCode);
}

/*
	Create()

	Crea la tabella (la apre azzerandola).
*/
int CBerkeleyDB::Create(void)
{
	int nCode = 0;

	if(!IsOpen())
		nCode = Open(DB_TRUNCATE);
	else
		nCode = SetLastError(DB_RETCODE_EALREDYOPEN);

	return(nCode);
}

/*
	SetLastError()

	Imposta il codice d'errore.
*/
int CBerkeleyDB::SetLastError(int nCode)
{
	// normalizza il codice di ritorno
	if(nCode < DB_OK)
	{
		switch(nCode)
		{
			case 0:
				nCode = DB_OK;
				break;
			case DB_NOTFOUND:
				nCode = DB_RETCODE_NOTFOUND;
				break;
			case DB_RUNRECOVERY:
				nCode = DB_RETCODE_EINVALIDHANDLE;
				break;
			case EAGAIN:
				nCode = DB_RETCODE_ELOCKFAILURE;
				break;
			case EINVAL:
				nCode = DB_RETCODE_EINVALIDPARAM;
				break;
			case ENOENT:
				nCode = DB_RETCODE_EINVALIDRESOURCE;
				break;
		}
	}

	// codice numerico
	m_Database.table.stat.errnum = nCode;
	
	// codice stringa
	_snprintf(m_Database.table.stat.errstr,sizeof(m_Database.table.stat.errstr)-1,"error code: %d",nCode);

	for(int i = 0; i < ARRAY_SIZE(db_error); i++)
	{
		if(db_error[i].code==nCode)
		{
			strcpyn(m_Database.table.stat.errstr,db_error[i].desc,sizeof(m_Database.table.stat.errstr));
			break;
		}
	}

#if defined(_WINDOWS)
	if(nCode >= DB_RETCODE_ERRORCODE)
		if(m_bShowErrors)
			::MessageBox(NULL,m_Database.table.stat.errstr,"CBerkeleyDB::SetLastError()",MB_ICONERROR|MB_SETFOREGROUND|MB_TOPMOST);
#endif

	return(nCode);
}

/*
	SetCursor()

	Imposta il cursore (indice) corrente.
*/
int CBerkeleyDB::SetCursor(int i)
{
	int nCode = 0;

	if(IsOpen())
	{
		if(m_Database.table.totindex > 0)
		{
			ASSERTEXPR(i >= 0 && i < m_Database.table.totindex);

			if(i >= 0 && i < m_Database.table.totindex)
			{
				m_pCurrentCursor = m_pIdxCursorArray[i];
				m_nCurrentCursorIndex = i;
			}
			else
				nCode = SetLastError(DB_RETCODE_EINVALIDINDEX);
		}
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	ResetRecord()

	Azzera i campi del record.
*/
void CBerkeleyDB::ResetRecord(void)
{
	if(IsOpen())
	{
		for(int i = 0; i < m_Database.table.totfield; i++)
			memset(m_Database.table.row[i].value,'\0',m_Database.table.row[i].size);
	}
	else
		SetLastError(DB_RETCODE_EINVALIDHANDLE);
}

/*
	BlankRecord()

	Svuota i campi del record.
*/
void CBerkeleyDB::BlankRecord(void)
{
	if(IsOpen())
	{
		for(int i = 0; i < m_Database.table.totfield; i++)
			memset(m_Database.table.row[i].value,' ',m_Database.table.row[i].size);
	}
	else
		SetLastError(DB_RETCODE_EINVALIDHANDLE);
}

/*
	GetFirst()

	Recupera la prima chiave.
*/
int CBerkeleyDB::GetFirst(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		DBC* pCursor = GetCursor();
		int nCursor = GetCursorIndex();

		// resetta la coppia chiave/dati
		ResetPair();

		// recupera la prima chiave con il cursore (indice) corrente
		if((nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_FIRST))==0)
		{
			// se il cursore corrente non e' la chiave primaria cerca il record con la chiave secondaria
			// (ossia cerca negli indici il record relativo al valore)
			if(nCursor!=GetPrimaryKeyCursorIndex())
				nCode = GetSecondaryKey(nCursor);
		}

		if(nCode!=0)
		{
			nCode = SetLastError(nCode);
			ResetRecord();
		}
		else
			MemoryToRecord();
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	GetNext()

	Recupera la chiave successiva.
*/
int CBerkeleyDB::GetNext(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		DBC* pCursor = GetCursor();
		int nCursor = GetCursorIndex();

		// resetta la coppia chiave/dati
		ResetPair();

		// recupera la chiave seguente con il cursore (indice) corrente
		if((nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_NEXT))==0)
		{
			// se il cursore corrente non e' la chiave primaria cerca il record con la chiave secondaria
			if(nCursor!=GetPrimaryKeyCursorIndex())
				nCode = GetSecondaryKey(nCursor);
		}

		if(nCode!=0)
		{
			nCode = SetLastError(nCode);
			ResetRecord();
		}
		else
			MemoryToRecord();
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	GetPrev()

	Recupera la chiave precedente.
*/
int CBerkeleyDB::GetPrev(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		DBC* pCursor = GetCursor();
		int nCursor = GetCursorIndex();

		// resetta la coppia chiave/dati
		ResetPair();

		// recupera la chiave precedente con il cursore (indice) corrente
		if((nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_PREV))==0)
		{
			// se il cursore corrente non e' la chiave primaria cerca il record con la chiave secondaria
			if(nCursor!=GetPrimaryKeyCursorIndex())
				nCode = GetSecondaryKey(nCursor);
		}

		if(nCode!=0)
		{
			nCode = SetLastError(nCode);
			ResetRecord();
		}
		else
			MemoryToRecord();
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	GetLast()

	Recupera l'ultima chiave.
*/
int CBerkeleyDB::GetLast(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		DBC* pCursor = GetCursor();
		int nCursor = GetCursorIndex();

		// resetta la coppia chiave/dati
		ResetPair();

		// recupera l'ultima chiave con il cursore (indice) corrente
		if((nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_LAST))==0)
		{
			// se il cursore corrente non e' la chiave primaria cerca il record con la chiave secondaria
			if(nCursor!=GetPrimaryKeyCursorIndex())
				nCode = GetSecondaryKey(nCursor);
		}

		if(nCode!=0)
		{
			nCode = SetLastError(nCode);
			ResetRecord();
		}
		else
			MemoryToRecord();
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	GetCurrent()

	Recupera la chiave corrente.
*/
int CBerkeleyDB::GetCurrent(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		DBC* pCursor = GetCursor();
		int nCursor = GetCursorIndex();

		// recupera la chiave corrente con il cursore (indice) corrente
		// assume che sia stata gia' chiamata PutKey() per impostare il valore della chiave corrente
		if((nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_SET))==0)
		{
			// se il cursore corrente non e' la chiave primaria cerca il record con la chiave secondaria
			if(nCursor!=GetPrimaryKeyCursorIndex())
				nCode = GetSecondaryKey(nCursor);
		}

		if(nCode!=0)
		{
			nCode = SetLastError(nCode);
			ResetRecord();
		}
		else
			MemoryToRecord();
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	GetSecondaryKey()

	Effettua la ricerca per chiave secondaria.
*/
int CBerkeleyDB::GetSecondaryKey(int index)
{
	int	nCode = -1;
	int	nFieldNum = -1;
	char	szField[MAX_KEYSIZE+1];
	char	szPrimaryKey[MAX_KEYSIZE+1],szSecondaryKey[MAX_KEYSIZE+1];
	int	nPrimaryKeyLen = -1,nSecondaryKeyLen = -1;
	DBC*	pCursor = GetPrimaryKeyCursor();

	// controlla che il numero dell'indice sia valido
	if(index >= 0 && index < m_Database.table.totindex)
		nFieldNum = m_Database.table.index[index].fieldnum;
	
	ASSERTEXPR(nFieldNum >= 0 && nFieldNum < m_Database.table.totfield);

	// controlla che il numero del campo sia valido
	if(nFieldNum >= 0 && nFieldNum < m_Database.table.totfield)
	{
		// la prima ricerca, quella che ha riempito key/data, e' stata effettuata con una delle chiavi
		// secondarie ottenendo in data.data la chiave primaria
		
		// recupera la chiave secondaria (presente in m_Key.data)
		nSecondaryKeyLen = m_Database.table.row[nFieldNum].size;
		ASSERTEXPR(nSecondaryKeyLen > 0 && nSecondaryKeyLen <= MAX_KEYSIZE);
		nSecondaryKeyLen = (nSecondaryKeyLen <= MAX_KEYSIZE) ? nSecondaryKeyLen : MAX_KEYSIZE;
		
		memset(szField,'\0',sizeof(szField));
		memcpy(szField,(char*)m_Key.data,nSecondaryKeyLen);
		
		memset(szSecondaryKey,' ',sizeof(szSecondaryKey));
		memcpy(szSecondaryKey,szField,strlen(szField));
		szSecondaryKey[nSecondaryKeyLen] = 0;

		// recupera la chiave primaria (presente in data.data)
		nPrimaryKeyLen = GetPrimaryKeySize();
		ASSERTEXPR(nPrimaryKeyLen > 0 && nPrimaryKeyLen <= MAX_KEYSIZE);
		nPrimaryKeyLen = (nPrimaryKeyLen <= MAX_KEYSIZE) ? nPrimaryKeyLen : MAX_KEYSIZE;

		memset(szField,'\0',sizeof(szField));
		memcpy(szField,(char*)m_Data.data,nPrimaryKeyLen);
		
		memset(szPrimaryKey,' ',sizeof(szPrimaryKey));
		memcpy(szPrimaryKey,szField,strlen(szField));
		szPrimaryKey[nPrimaryKeyLen] = 0;

		// imposta la coppia dati/dimensione per la chiave
		ResetPair();
		m_Key.data = szPrimaryKey;
		m_Key.size = nPrimaryKeyLen;

		// cerca, con l'indice primario, il record con chiave primaria uguale a quella ricavata con la
		// ricerca effettuata (nel chiamante) con l'indice secondario
		// (in altre parole assume che la coppia m_Key/data sia stata riempita dal chiamante con una ricerca
		// per chiave secondaria)
		// non carica/azzera il record (deve farlo il chiamante al ritorno)
		// non effettua nessun ciclo perche' in teoria la chiave primaria non puo' esistere duplicata
		nCode = pCursor->c_get(pCursor,&m_Key,&m_Data,DB_SET);
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDFIELDNUMBER);

	return(nCode);
}

/*
	Insert()

	Inserisce la chiave.
*/
int CBerkeleyDB::Insert(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		char* pPrimaryKeyData   = GetPrimaryKeyData();
		int   nPrimaryKeySize   = GetPrimaryKeySize();
		DBC*  pPrimaryKeyCursor = GetPrimaryKeyCursor();

		// azzera la struttura key/data
		ResetPair();

		// genera la chiave primaria per l'inserimento
		if(m_Database.table.totindex > 0)
		{
			int	nLastKeyLen;
			long	nLastKeyNum;
			char	szBuffer[MAX_PRIMARYKEY_SIZE*2];
			
			// chiave = chiave corrente + 1
			nLastKeyNum = atol(m_szLastKey) + 1L;
			memset(m_szLastKey,'0',sizeof(m_szLastKey));
			nLastKeyLen = strlen(ltoa(nLastKeyNum,szBuffer,MAX_PRIMARYKEY_SIZE));
			memcpy(m_szLastKey + (nPrimaryKeySize-nLastKeyLen),szBuffer,nLastKeyLen);
			m_szLastKey[nPrimaryKeySize] = '\0';
			memcpy(pPrimaryKeyData,m_szLastKey,nPrimaryKeySize);
		}
		
		// passa la chiave primaria nella struttura (key)
		m_Key.data = pPrimaryKeyData;
		m_Key.size = nPrimaryKeySize;

		// passa i dati nella struttura (data)
		RecordToMemory();

		// inserisce il record corrente nel database
		// considera come record corrente quello impostato dal chiamante con PutField()
		// notare che la chiave primaria non entra a formar parte dei dati
		nCode = pPrimaryKeyCursor->c_put(pPrimaryKeyCursor,&m_Key,&m_Data,DB_KEYLAST);

		// aggiorna gli indici inserendo un record con:
		// chiave	-> il campo relativo dell'indice secondario
		// dati	-> la chiave primaria
		if(nCode==0)
		{
			int	nRet;
			char	szField[MAX_KEYSIZE+1];
			char	szBuffer[MAX_KEYSIZE+1];
			int	nFieldNum,nFieldSize;

			// per ognuno degli indici della tabella
			for(int i = 0; i < m_Database.table.totindex; i++)
			{
				// ricava il numero e la dimensione del campo
				nFieldNum  = m_Database.table.index[i].fieldnum;
				ASSERTEXPR(nFieldNum >= 0 && nFieldNum < MAX_FIELDCOUNT);
				nFieldSize = m_Database.table.row[nFieldNum].size;
				ASSERTEXPR(nFieldSize > 0 && nFieldSize <= MAX_KEYSIZE);
				nFieldSize = nFieldSize > MAX_KEYSIZE ? MAX_KEYSIZE : nFieldSize;

				// ricava il valore del campo
				memset(szBuffer,'\0',sizeof(szBuffer));
				memset(szField,' ',sizeof(szField));
				memcpy(szBuffer,m_Database.table.row[nFieldNum].value,nFieldSize);
				szBuffer[nFieldSize] = '\0';
				memcpy(szField,szBuffer,strlen(szBuffer));

				// imposta la chiave (valore del campo usato come indice secondario)
				m_Key.data = szField;
				m_Key.size = nFieldSize;

				// imposta i dati (la chiave primaria)
				m_Data.data = pPrimaryKeyData;
				m_Data.size = nPrimaryKeySize;

				// inserisce il record nell'indice secondario (campo + chiave primaria)
				if((nRet = m_pIdxCursorArray[i]->c_put(m_pIdxCursorArray[i],&m_Key,&m_Data,DB_KEYLAST))!=0)
					nCode = nRet;
			}
		}

		if(nCode!=0)
			nCode = SetLastError(nCode);
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	Delete()

	Elimina la chiave.
*/
int CBerkeleyDB::Delete(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		char* pPrimaryKeyData   = GetPrimaryKeyData();
		int   nPrimaryKeySize   = GetPrimaryKeySize();
		DBC*  pPrimaryKeyCursor = GetPrimaryKeyCursor();

		// azzera la struttura key/data
		ResetPair();

		// passa la chiave primaria nella struttura (key)
		m_Key.data = pPrimaryKeyData;
		m_Key.size = nPrimaryKeySize;

		// passa i dati nella struttura (data)
		RecordToMemory();

#if 0
		// non considera il record relativo alla ultima get ma (ri)cerca il record impostato con PutField()
		// e se lo trova lo elimina
		nCode = pPrimaryKeyCursor->c_get(pPrimaryKeyCursor,&m_Key,&m_Data,DB_SET);
#else
		// elimina il record corrente dal database
		// considera come record corrente quello impostato dal chiamante con PutField()
		nCode = pPrimaryKeyCursor->c_del(pPrimaryKeyCursor,0);
#endif

		// aggiorna gli indici
		// cerca il primo record relativo alla chiave secondaria (nell'indice secondario possono esistere
		// chiavi duplicate) e una volta trovato cerca il record che contenga (come dati) la chiave primaria
		if(nCode==0)
		{
			char	szBuffer[MAX_KEYSIZE+1];
			char	szPrimaryKey[MAX_KEYSIZE+1];
			char	szField[MAX_KEYSIZE+1];
			int	nFieldNum,nFieldSize;

			memset(szBuffer,'\0',sizeof(szBuffer));
			memset(szPrimaryKey,' ',sizeof(szPrimaryKey));
			memcpy(szBuffer,pPrimaryKeyData,nPrimaryKeySize);
			szBuffer[nPrimaryKeySize] = '\0';
			memcpy(szPrimaryKey,szBuffer,strlen(szBuffer));

			for(int i = 0; i < m_Database.table.totindex; i++)
			{
				memset(szBuffer,'\0',sizeof(szBuffer));
				memset(szField,' ',sizeof(szField));

				// ricava il nome e dimensione del campo
				nFieldNum  = m_Database.table.index[i].fieldnum;
				ASSERTEXPR(nFieldNum >= 0 && nFieldNum < MAX_FIELDCOUNT);
				nFieldSize = m_Database.table.row[nFieldNum].size;
				ASSERTEXPR(nFieldSize > 0 && nFieldSize <= MAX_KEYSIZE);

				// ricava il valore del campo
				memcpy(szBuffer,m_Database.table.row[nFieldNum].value,nFieldSize);
				szBuffer[nFieldSize] = '\0';
				memcpy(szField,szBuffer,strlen(szBuffer));
				
				// imposta la chiave con il valore del campo
				m_Key.data = szField;
				m_Key.size = nFieldSize;

				// cerca la prima chiave relativa al campo
				if((nCode = m_pIdxCursorArray[i]->c_get(m_pIdxCursorArray[i],&m_Key,&m_Data,DB_SET))==0)
				{
					// scorre i record trovati cercando quello con chiave primaria uguale a quella impostata sopra
					do {
						if(memcmp(szPrimaryKey,(char*)m_Data.data,nPrimaryKeySize)==0)
						{
							// elimina il record dall'indice secondario
							m_pIdxCursorArray[i]->c_del(m_pIdxCursorArray[i],0);
							break;
						}
					}
					while((nCode = m_pIdxCursorArray[i]->c_get(m_pIdxCursorArray[i],&m_Key,&m_Data,DB_NEXT))==0);
				}
			}
		}

		if(nCode!=0)
			nCode = SetLastError(nCode);
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	Replace()

	Sostituisce il record corrente.
*/
int CBerkeleyDB::Replace(void)
{
	int nCode = 0;

	if(IsOpen())
	{
		char* pPrimaryKeyData   = GetPrimaryKeyData();
		int   nPrimaryKeySize   = GetPrimaryKeySize();
		DBC*  pPrimaryKeyCursor = GetPrimaryKeyCursor();

		// azzera la struttura key/data
		ResetPair();

		// passa la chiave primaria nella struttura (key)
		m_Key.data = pPrimaryKeyData;
		m_Key.size = nPrimaryKeySize;

		// passa i dati nella struttura (data)
		RecordToMemory();

		// sostituisce il record corrente del database (considera come record corrente quello impostato
		// dal chiamante con con PutField())
		//
		// notare che NON si possono sostituire campi che siano chiave: l'aggiornamento si limita al
		// database e non include gli indici perche' quando viene chiamata la PutField() della classe
		// derivata viene cambiato il valore del campo chiave per cui la successiva ricerca sugli indici
		// fallirebbe
		//
		// affinche la replace funzioni la tabella deve prevedere perlomeno un indice (secondario)
		if((nCode = pPrimaryKeyCursor->c_put(pPrimaryKeyCursor,&m_Key,&m_Data,DB_CURRENT))!=0)
			nCode = SetLastError(nCode);
	}
	else
		nCode = SetLastError(DB_RETCODE_EINVALIDHANDLE);

	return(nCode);
}

/*
	PutKey()

	Imposta il valore per la chiave.
*/
void CBerkeleyDB::PutKey(char *pValue,int nIndex)
{
	// da mem a key(.data/.size)
	// imposta il valore della chiave (relativa all'indice) per la ricerca
	static char szBuffer[MAX_KEYSIZE+1];
	int nFieldNum = -1;
	int nKeyLen,nKeySize;

	ResetKeyPair();

	// se la tabella non possiede indici assume il primo campo (base 0)
	if(m_Database.table.totindex > 0)
	{
		if(nIndex >= 0 && nIndex < m_Database.table.totindex)
			nFieldNum = m_Database.table.index[nIndex].fieldnum;
		ASSERTEXPR(nFieldNum >= 0 && nFieldNum < MAX_FIELDCOUNT);
	}
	else
		nFieldNum = 0;

	// controlla che il numero del campo sia valido
	if(nFieldNum >= 0 && nFieldNum < m_Database.table.totfield)
	{
		// imposta i valori (dati/dimensione) per la chiave
		nKeySize = m_Database.table.row[nFieldNum].size;
		ASSERTEXPR(nKeySize > 0 && nKeySize <= MAX_KEYSIZE);
		nKeySize = nKeySize > 0 && nKeySize <= MAX_KEYSIZE ? nKeySize : MAX_KEYSIZE;
		nKeyLen = strlen(pValue);
		nKeyLen = nKeyLen > nKeySize ? nKeySize : nKeyLen;

		memset(szBuffer,' ',sizeof(szBuffer));
		memcpy(szBuffer,pValue,nKeyLen);
		szBuffer[sizeof(szBuffer)-1] = '\0';

		m_Key.data = szBuffer;
		m_Key.size = nKeySize;
	}
	else
		SetLastError(DB_RETCODE_EINVALIDFIELDNUMBER);
}

/*
	PutField()

	Imposta il valore per il campo.
*/
void CBerkeleyDB::PutField(int nFieldNum,const char* pValue)
{
	// da mem a rec->field
	if(nFieldNum >= 0 && nFieldNum < m_Database.table.totfield)
	{
		ASSERTEXPR(m_Database.table.row[nFieldNum].size <= MAX_FIELDSIZE);
		
		int len  = strlen(pValue);
		len = len > m_Database.table.row[nFieldNum].size ? m_Database.table.row[nFieldNum].size : len;

		// passa il contenuto del buffer nel campo del record
		// (prima di effettuare la copia formatta il campo del record a spazi)
		memset(m_Database.table.row[nFieldNum].value,' ',m_Database.table.row[nFieldNum].size);
		memcpy(m_Database.table.row[nFieldNum].value,pValue,len);
	}
}

/*
	GetField()

	Ricava il valore del campo.
*/
const char* CBerkeleyDB::GetField(int nFieldNum)
{
	// da rec->field a mem
	static char szField[MAX_FIELDSIZE+1];

	if(nFieldNum >= 0 && nFieldNum < m_Database.table.totfield)
	{
		// passa il contenuto del campo del record (formattato a spazi) nel buffer, terminandolo con '\0'
		memset(szField,'\0',sizeof(szField));
		ASSERTEXPR(m_Database.table.row[nFieldNum].size <= MAX_FIELDSIZE);
		memcpy(szField,m_Database.table.row[nFieldNum].value,m_Database.table.row[nFieldNum].size);
	}

	return(szField);
}

/*
	MemoryToRecord()

	Passa i dati dal buffer interno al record.
*/
void CBerkeleyDB::MemoryToRecord(void)
{
	// passa i dati dal campo data.data al record
	char* p = (char*)m_Data.data;

	// azzera il record
	ResetRecord();

	if(p && *p)
	{
		int i = 0,ofs = 0;

		// se la tabella prevede indici, copia la chiave primaria nel primo campo de record
		if(m_Database.table.totindex > 0)
		{
			memcpy(m_Database.table.row[0].value,(char*)m_Key.data,m_Database.table.row[0].size);
			i++;
			ofs = m_Database.table.row[0].size;
		}

		// passa il contenuto del buffer nei campi del record
		// (notare che i valori presenti nel buffer sono gia' formattati a spazio)
		for(; i < m_Database.table.totfield; i++)
			memcpy(m_Database.table.row[i].value,p + m_Database.table.row[i].ofs - ofs,m_Database.table.row[i].size);
	}
}

/*
	RecordToMemory()

	Passa i dati dal record al buffer interno.
*/
void CBerkeleyDB::RecordToMemory(void)
{
	int i,nSize,nFieldSize;

	// azzera il buffer interno per i dati
	memset(m_pDataBuffer,'\0',m_nDataBufferSize);

	// se la tabella prevede indici salta la chiave primaria (non la include nei dati)
	i = m_Database.table.totindex > 0 ? 1 : 0;

	// passa i campi del record nel buffer interno (i valori sono gia' formattati a spazio)
	for(nSize = 0; i < m_Database.table.totfield; i++)
	{
		nFieldSize = m_Database.table.row[i].size;
		nFieldSize = (nFieldSize < (m_nDataBufferSize - 1 - nSize)) ? nFieldSize : (m_nDataBufferSize - 1 - nSize);

		memcpy(	m_pDataBuffer+nSize,
				m_Database.table.row[i].value,
				nFieldSize
				);

		nSize += nFieldSize;
	}

	// passa il contenuto del buffer interno nella struttura
	ResetDataPair();
	m_Data.data = m_pDataBuffer;
	m_Data.size = nSize;
}
