/*
	CBase.h
	Classe derivata per interfaccia con Berkeley DB (definisce l'oggetto).
	Luca Piergentili, 04/11/99
	lpiergentili@yahoo.com
	http://www.geocities.com/lpiergentili/
*/
#ifndef _CBASE_H
#define _CBASE_H 1

#ifdef _WINDOWS
  #include "window.h"
  #ifdef WIN32_MFC
    #include "CMutexEx.h"
  #endif
#else
  #include "typedef.h"
#endif
#include "CBerkeleyDB.h"
#include "CDateTime.h"

/*
	CBASE_FLAG_...
	flags per filtri sui campi
*/
#define CBASE_FLAG_NONE			0x0000	// nessun azione
#define CBASE_FLAG_TOUPPER		0x0002	// a -> A
#define CBASE_FLAG_TOLOWER		0x0004	// A -> a
#define CBASE_FLAG_NUMBER		0x0008	// 0-9
#define CBASE_FLAG_CHAR			0x0010	// a-z
#define CBASE_FLAG_PUNCT			0x0020	// .,;:?!&%$#@(){}[]-+*/
#define CBASE_FLAG_USERDEFINED	0x0040	// definito dall'utente
#define CBASE_MAX_NUMBER_PICTURE	32		// dimensioni dei buffer per la picture
#define CBASE_MAX_CHAR_PICTURE	128
#define CBASE_MAX_PUNCT_PICTURE	128
#define CBASE_MAX_USER_PICTURE	128

/*
	CBASE_TABLE
	struttura per la definizione della tabella
*/
struct CBASE_TABLE {
	char* field;		// nome campo
	char type;		// tipo
	int size;			// dimensione
	int dec;			// decimali
};

/*
	CBASE_INDEX
	struttura per la definizione dell'indice
*/
struct CBASE_INDEX {
	char* file;		// nome file
	char* name;		// nome indice
	char* field;		// nome campo associato
};

class CBase : public CBerkeleyDB
{
public:
	CBase(const char*,CBASE_TABLE*,CBASE_INDEX*);
	virtual ~CBase();

	// oggetto interno
	inline BOOL	IsValid			(void)				{return(m_bIsValid);}
	inline void	SetValid			(BOOL flag)			{m_bIsValid = flag;}
	
	// database
	BOOL			Open				(void);
	BOOL			Zap				(void);
	void			Close			(void);

	inline BOOL	Bof				(void)				{return(m_pDatabase->table.stat.bof);}
	inline BOOL	Eof				(void)				{return(m_pDatabase->table.stat.eof);}

	BOOL			GetCurrent		(int nIndex = -1);
	BOOL			GetFirst			(int nIndex = -1);
	inline BOOL	GoTop			(int nIndex  = -1)		{return(CBase::GetFirst(nIndex));}
	BOOL			GetNext			(int nIndex = -1);
	BOOL			GetPrev			(int nIndex = -1);
	BOOL			GetLast			(int nIndex = -1);
	inline BOOL	GoBottom			(int nIndex = -1)		{return(CBase::GetLast(nIndex));}

	int			GetFieldCount		(void);
	const char*	GetFieldNames		(void);
	int			GetFieldNumberByName(const char* pFieldName);

	inline int	GetIndexCount		(void)				{return(CBase::m_pDatabase->table.totindex);}
	const char*	GetIndexNames		(void);
	int			GetIndexNumberByName(const char* pIndexName);
	
	int			SetIndex			(int nIndex);
	BOOL			Seek				(const int nValue,int nIndex = -1);
	BOOL			Seek				(const char* pValue,int nIndex = -1);
	BOOL			Find				(const char* pValue);

	inline BOOL	Insert			(void)				{return(CBerkeleyDB::Insert()==0);}
	inline BOOL	Delete			(void)				{return(CBerkeleyDB::Delete()==0);}
	inline BOOL	Replace			(void)				{return(CBerkeleyDB::Replace()==0);}

	void			PutField			(int nFieldNum,const char* pValue);
	void			PutField			(int nFieldNum,BOOLEAN bValue);
	void			PutField			(int nFieldNum,CDateTime* pDateTime);
	void			PutField			(int nFieldNum,CDateTime& pDateTime);
	void			PutField			(int nFieldNum,int nValue);
	void			PutField			(int nFieldNum,long nValue);
	void			PutField			(int nFieldNum,unsigned long nValue);
	void			PutField			(int nFieldNum,float nValue);

	const char*	GetField			(int nFieldNum);
	char*		GetField			(int nFieldNum,char* pBuffer,int nSize);
	BOOLEAN		GetField			(int nFieldNum,BOOLEAN& bValue);
	const char*	GetField			(int nFieldNum,CDateTime* pDateTime);
	const char*	GetField			(int nFieldNum,CDateTime& pDateTime);
	int			GetField			(int nFieldNum,int& nValue);
	long			GetField			(int nFieldNum,long& nValue);
	unsigned long	GetField			(int nFieldNum,unsigned long& nValue);
	float		GetField			(int nFieldNum,float& nValue);

	// picture
	BOOL			SetNumberPicture	(const char* pPicture);
	BOOL			SetCharPicture		(const char* pPicture);
	BOOL			SetPunctPicture	(const char* pPicture);
	BOOL			SetUserPicture		(const char* pPicture);
	void			SetFieldFlags		(int nFieldNum,unsigned long nFlags);

	// errori
	inline int	GetLastError		(void)				{return(CBerkeleyDB::GetLastError());}
	inline const char* GetLastErrorString(void)				{return(CBerkeleyDB::GetLastErrorString());}

	// import/export
	int			Load				(const char* pFileName);
	int			Dump				(const char* pFileName,BOOL bDumpPrimaryKey = FALSE);

protected:
	// puntatore al buffer per il record
	char*		m_pRecordBuffer;

private:
	// tipi numerici
	union NUMBER {
		int			integer;	// intero
		long			longint;	// lungo
		unsigned long	ulongint;	// esageratamente lungo
		float		floatnum;	// virgola mobile
	};

	enum NUMTYPE {
		integer_type,
		longint_type,
		ulongint_type,
		floatnum_type
	};

	void	PutFieldNumber			(int,NUMBER,NUMTYPE);
	void	GetFieldNumber			(int,NUMBER&,NUMTYPE);

	// picture
	char* m_pPictureNumber;
	char* m_pPictureChar;
	char* m_pPicturePunct;
	char* m_pPictureUserDefined;
	
	void	SetDefaultNumberPicture	(void);
	void	SetDefaultCharPicture	(void);
	void	SetDefaultPunctPicture	(void);
	void	SetDefaultUserPicture	(void);
	void	SetFieldFormat			(char* pField,int nSize,unsigned long nFlags);
	void	SetFieldFormatByPicture	(char* pField,int nSize,unsigned long nFlags);
	
	// puntatore alla definizione della tabella (della classe base)
	DATABASE* m_pDatabase;

	// flag per esito costruttore
	BOOL m_bIsValid;
};

#endif // _CBASE_H
