/*
	CBerkeleyDB.h
	Classe base per interfaccia con Berkeley DB (usare solo come virtuale pura).
	Luca Piergentili, 04/11/99
	lpiergentili@yahoo.com
	http://www.geocities.com/lpiergentili/
*/
#ifndef _CBERKELEYDB_H
#define _CBERKELEYDB_H 1

#include <string.h>
#ifdef _WINDOWS
  #include "window.h"
  #ifdef WIN32_MFC
    #include "CMutexEx.h"
  #endif
#else
  #include "typedef.h"
#endif
#include "db.h"

// crea la referenza alla DLL
#ifdef _DEBUG
  #pragma comment(lib,"BerkeleyDB.d.lib")
#else
  #pragma comment(lib,"BerkeleyDB.lib")
#endif

// generiche
#define MAX_ERRORCODE_STRING	128			// dim. stringa x messaggio d'errore
#define MAX_PRIMARYKEY_SIZE	10			// dim. max per la chiave primaria
#define MAX_KEYSIZE			1024			// dim. max per la chiave
#define MAX_FIELDSIZE		MAX_KEYSIZE	// dim. max per il campo del record
#define MAX_FIELDCOUNT		128			// numero max. di campi x tabella

/*
	ROW
	struttura per la definizione del campo del record
*/
struct ROW {
	int		num;					// numero del campo nel record (progressivo a base 0)
	int		ofs;					// offset del campo nel record (a base 0)
	char*	name;				// nome del campo
	char		type;				// tipo del campo
	int		size;				// dimensione del campo
	int		dec;					// decimali del campo
	char*	value;				// contenuto del campo (il buffer relativo e' l'istanza della struct per il record)
	unsigned long flags;			// flag per operazioni (filtri) sul campo
};

/*
	INDEX
	struttura per la definizione dell'indice
*/
struct INDEX {
	char		filename[_MAX_PATH+1];	// nome file indice
	char*	name;				// nome indice
	char*	fieldname;			// nome del campo della tabella utilizzato come chiave
	int		fieldnum;				// numero del campo della tabella utilizzato come chiave (progressivo a base 0)
};

/*
	TABLE_STAT
	struttura per lo status attuale della tabella
*/
struct TABLE_STAT {
	int		bof;					// flag per inizio file
	int		eof;					// flag per fine file
	int		errnum;				// codice numerico ultimo errore
	char		errstr[MAX_ERRORCODE_STRING+1]; // descrizione ultimo errore
};

/*
	TABLE
	struttura per la definizione della tabella
*/
struct TABLE {
	char		filename[_MAX_PATH+1];	// nome tabella
	int		totfield;				// numero totale di campi
	ROW*		row;					// array per la definizione del record, da allocare a run-time
	int		totindex;				// numero totale indici
	INDEX*	index;				// array per la definizione degli indici, da allocare a run-time
	TABLE_STAT stat;				// status corrente
};

/*
	DATABASE
	struttura per la definizione del database
*/
struct DATABASE {
	TABLE	table;				// definizione della tabella
};

/*
	DB_...
	codici di ritorno:
	
	1000		info
	2000		status
	>=3000	errori
*/
// 1000 (info)
#define DB_OK						1000
// 2000 (status)
#define DB_RETCODE_STATUSCODE			2000	// base per test
#define DB_RETCODE_NOTFOUND			2000
#define DB_RETCODE_BOF				2001
#define DB_RETCODE_EOF				2002
// 3000 (errori)
#define DB_RETCODE_ERRORCODE			3000 // base per test
#define DB_RETCODE_EOPENTABLE			3000
#define DB_RETCODE_EOPENCURSOR		3001
#define DB_RETCODE_EOPENINDEX			3002
#define DB_RETCODE_EOPENIDXCURSOR		3003
#define DB_RETCODE_ECREATETABLE		3004
#define DB_RETCODE_ECLOSETABLE		3005
#define DB_RETCODE_ECLOSECURSOR		3006
#define DB_RETCODE_ECLOSEINDEX		3007
#define DB_RETCODE_ECLOSEIDXCURSOR		3008
#define DB_RETCODE_EALREDYOPEN		3009
#define DB_RETCODE_EINVALIDHANDLE		3010
#define DB_RETCODE_ELOCKFAILURE		3011
#define DB_RETCODE_EINVALIDPARAM		3012
#define DB_RETCODE_EINVALIDRESOURCE	3013
#define DB_RETCODE_EINVALIDINDEX		3014
#define DB_RETCODE_ETABLENOTOPEN		3015
#define DB_RETCODE_EINVALIDFIELDNUMBER	3016
#define DB_RETCODE_EINVALIDFIELDSIZE	3017
#define DB_RETCODE_EALLOCFAILURE		3018
#define DB_RETCODE_EOPENFAILURE		3019
// >3000 (errori)
#define DB_RETCODE_ERROR				4000
#define DB_RETCODE_UNKNOWERROR		6666

class CBerkeleyDB
{
public:
	CBerkeleyDB();
	virtual ~CBerkeleyDB();
	
protected:
	int			Open					(u_int32_t = DB_CREATE);
	int			Close				(void);
	int			Create				(void);
	inline int	IsOpen				(void)				{return(m_pDbHandle!=NULL);}

	int			SetLastError			(int);
	inline void	ResetLastError			(void)				{m_Database.table.stat.errnum = 0,memset(&(m_Database.table.stat.errstr),'\0',sizeof(m_Database.table.stat.errstr));}
	inline int	GetLastError			(void)				{return(m_Database.table.stat.errnum);}
	inline const char* GetLastErrorString	(void)				{return((const char*)m_Database.table.stat.errstr);}

#if defined(_WINDOWS)
	inline void	ShowErrors			(BOOL bFlag)			{m_bShowErrors = bFlag;}
#endif

	int			SetCursor				(int);
	inline DBC*	GetCursor				(void)				{return(m_pCurrentCursor);}
	inline int	GetCursorIndex			(void)				{return(m_nCurrentCursorIndex);}
	inline DBC*	GetPrimaryKeyCursor		(void)				{return(m_pPrimaryKeyCursor);}
	inline int	GetPrimaryKeyCursorIndex	(void)				{return(m_nPrimaryKeyCursorIndex);}

	inline char*	GetPrimaryKeyData		(void)				{return(m_Database.table.row[0].value);}
	inline int	GetPrimaryKeySize		(void)				{return(m_Database.table.row[0].size);}

	inline void	ResetKeyPair			(void)				{memset(&m_Key,'\0',sizeof(m_Key));}
	inline void	ResetDataPair			(void)				{memset(&m_Data,'\0',sizeof(m_Data));}
	inline void	ResetPair				(void)				{ResetKeyPair(),ResetDataPair();}
	void			ResetRecord			(void);
	void			BlankRecord			(void);
	
	int			GetFirst				(void);
	int			GetNext				(void);
	int			GetPrev				(void);
	int			GetLast				(void);
	int			GetCurrent			(void);
	int			GetSecondaryKey		(int);

	int			Insert				(void);
	int			Delete				(void);
	int			Replace				(void);

	void			PutKey				(char* pValue,int nIndex);

	void			PutField				(int nFieldNum,const char* pValue);
	const char*	GetField				(int nFieldNum);
	void			MemoryToRecord			(void);
	void			RecordToMemory			(void);

	DATABASE*		GetDatabase			(void)				{return(&m_Database);}

private:
	DATABASE		m_Database;						// struct per la tabella
	char*		m_pDataBuffer;						// buffer per il campo data.data per i dati (campi del record)
	int			m_nDataBufferSize;					// dimensione
	int			m_nRecordSize;						// dimensione del record (somma dei campi)
	DB_ENV		m_dbenv;							// (db)
	DB_INFO		m_dbinfo;							// (db)
	DBT			m_Key,m_Data;						// coppia chiave/valore (db)
	DB*			m_pDbHandle;						// handle della tabella (db)
	DBC*			m_pDbCursor;						// cursore per vista sulla tabella (db)
	DBC*			m_pCurrentCursor;					// cursore corrente
	int			m_nCurrentCursorIndex;				// indice del cursore corrente
	DBC*			m_pPrimaryKeyCursor;				// cursore per la chiave primaria
	int			m_nPrimaryKeyCursorIndex;			// indice del cursore per la chiave primaria
	DB**			m_pIdxHandleArray;					// array per gli handles degli indici
	DBC**		m_pIdxCursorArray;					// array per i cursori degli indici
	char			m_szLastKey[(MAX_PRIMARYKEY_SIZE*2)+1];	// buffer per la generazione della chiave primaria
#if defined(_WINDOWS)
	BOOL			m_bShowErrors;						// flag per visualizzazione errori
#endif
};

#endif // _CBERKELEYDB_H
