/*
	CTable.h
	Classe base da cui derivare le tabelle per l'accesso al database.
	Non deriva da CBase ma incorpora l'oggetto per gestire il nome della tabella.
	Include il mutex e le relative funzioni di lock solo per MFC.
	Luca Piergentili, 14/07/99
	lpiergentili@yahoo.com
	http://www.geocities.com/lpiergentili/
*/
#ifndef _CTABLE_H
#define _CTABLE_H 1

#ifdef _WINDOWS
  #include "window.h"
  #ifdef WIN32_MFC
    #include "CMutexEx.h"
  #endif
#else
  #include "typedef.h"
#endif
#include "CBase.h"

class CTable
{
public:
	// costruttore/distruttore
#ifdef WIN32_MFC
	CTable()	{m_pCBase = NULL,m_TableMutex = NULL;}
	~CTable()	{if(m_pCBase) delete m_pCBase,m_pCBase = NULL; if(m_TableMutex) delete m_TableMutex,m_TableMutex = NULL;}
#else
	CTable()	{m_pCBase = NULL;}
	~CTable()	{if(m_pCBase) delete m_pCBase,m_pCBase = NULL;}
#endif

	// virtuali pure
	virtual const char*			GetClassName		(void) = 0;
	virtual const char*			GetStaticTableName	(void) = 0;
	virtual const char*			GetTableName		(void) = 0;
	virtual const char*			GetTablePathName	(void) = 0;
	virtual const CBASE_TABLE*	GetTableStruct		(void) = 0;
	virtual const CBASE_INDEX*	GetIndexStruct		(void) = 0;
	virtual const int			GetRecordLength	(void) = 0;
	virtual const char*			GetRecordAsString	(void) = 0;
	virtual void				ResetMemvars		(void) = 0;
	virtual void				GatherMemvars		(void) = 0;
	virtual void				ScatterMemvars		(BOOL = TRUE) = 0;

	// CBase
	inline BOOL				Open				(void)						{return(m_pCBase->Open());}
	inline BOOL				Zap				(void)						{return(m_pCBase->Zap());}
	inline void				Close			(void)						{m_pCBase->Close();}
	inline BOOL				Bof				(void)						{return(m_pCBase->Bof());}
	inline BOOL				Eof				(void)						{return(m_pCBase->Eof());}
	inline BOOL				GetCurrent		(int n = -1)					{return(m_pCBase->GetCurrent(n));}
	inline BOOL				GetFirst			(int n = -1)					{return(m_pCBase->GetFirst(n));}
	inline BOOL				GoTop			(int n = -1)					{return(m_pCBase->GoTop(n));}
	inline BOOL				GetNext			(int n = -1)					{return(m_pCBase->GetNext(n));}
	inline BOOL				GetPrev			(int n = -1)					{return(m_pCBase->GetPrev(n));}
	inline BOOL				GetLast			(int n = -1)					{return(m_pCBase->GetLast(n));}
	inline BOOL				GoBottom			(int n = -1)					{return(m_pCBase->GoBottom(n));}
	inline int				GetFieldCount		(void)						{return(m_pCBase->GetFieldCount());}
	inline const char*			GetFieldNames		(void)						{return(m_pCBase->GetFieldNames());}
	inline int				GetFieldNumberByName(const char* name)				{return(m_pCBase->GetFieldNumberByName(name));}
	inline int				GetIndexCount		(void)						{return(m_pCBase->GetIndexCount());}
	inline const char*			GetIndexNames		(void)						{return(m_pCBase->GetIndexNames());}
	inline int				GetIndexNumberByName(const char* name)				{return(m_pCBase->GetIndexNumberByName(name));}
	inline int				SetIndex			(int n)						{return(m_pCBase->SetIndex(n));}
	inline BOOL				Seek				(const int value,int n = -1)		{return(m_pCBase->Seek(value,n));}
	inline BOOL				Seek				(const char* value,int n = -1)	{return(m_pCBase->Seek(value,n));}
	inline BOOL				Find				(const char* value)				{return(m_pCBase->Find(value));}
	inline BOOL				Insert			(void)						{return(m_pCBase->Insert());}
	inline BOOL				Delete			(void)						{return(m_pCBase->Delete());}
	inline BOOL				DeleteSynchronize	(void)						{m_pCBase->Delete(); m_pCBase->Close(); return(m_pCBase->Open());}
	inline BOOL				Replace			(void)						{return(m_pCBase->Replace());}
	inline BOOL				Synchronize		(void)						{m_pCBase->Close(); return(m_pCBase->Open());}
	inline int				Load				(const char* s)				{return(m_pCBase->Load(s));}
	inline int				Dump				(const char* s,BOOL f = FALSE)	{return(m_pCBase->Dump(s,f));}

	// lock
#ifdef WIN32_MFC
	BOOL						LockTable			(unsigned int timeout = MUTEX_LOCK_TIMEOUT);
	BOOL						UnlockTable		(void);
#else
	inline BOOL				LockTable			(unsigned int timeout = 30) {return(TRUE);}
	inline BOOL				UnlockTable		(void) {return(TRUE);}
#endif

private:
	// mutex
#ifdef WIN32_MFC
	BOOL						Lock				(unsigned int timeout = MUTEX_LOCK_TIMEOUT);
	BOOL						Unlock			(void);
#else
	inline BOOL				Lock				(unsigned int timeout = 30) {return(TRUE);}
	inline BOOL				Unlock			(void) {return(TRUE);}
#endif

protected:
	// oggetto per la tabella (da costruire nella classe derivata)
	CBase* m_pCBase;

#ifdef WIN32_MFC
	// oggetto per il lock della tabella (costruito dalla classe derivata)
	// il mutex non puo' essere anonimo, ma deve essere nominato con il nome della tabella
	CMutexEx* m_TableMutex;
#endif
};

#endif // _CTABLE_H
